'''
Help function for the markov walk.
'''
import numpy as np
from ugd.help_function.util import get_path, form_to_set_index


def update_violation_matrix(graph, cycle_start_node, active_cyclenode, ind, violation_matrix):
    # counts the number of changes between/within set partition
    if cycle_start_node == None:
        return violation_matrix
    else:
        path = get_path(startnode=cycle_start_node, active_start=active_cyclenode, graph=graph, pathnumber=ind)
        is_aktive = active_cyclenode  # change, start aktie or passive depending on whether path starts with an aktiv edge
        for edge in iter_edge_in_path(path):
            violation_matrix = update_violation_matrix_by_edge(graph, violation_matrix, edge, is_aktive)
            is_aktive = not (is_aktive)
        return violation_matrix


def update_violation_matrix_by_edge(graph, violation_matrix, edge, is_active):
    # determines the changes of a single edge for the violation matrix
    from_id, to_id = form_to_set_index(graph, edge)
    if graph.is_directed:
        if is_active:
            violation_matrix[from_id, to_id] -= 1
        else:
            violation_matrix[to_id, from_id] += 1
    else:
        if is_active:
            violation_matrix[from_id, to_id] -= 1
            violation_matrix[to_id, from_id] -= 1
        else:
            violation_matrix[from_id, to_id] += 1
            violation_matrix[to_id, from_id] += 1

    return violation_matrix


def iter_edge_in_path(path):
    n = path.__len__()
    for i in range(n - 1):
        yield (path[i], path[i + 1])


def switch_cycles(graph, cycle_nodes, active_startnodes):
    for ind, cycle_node in enumerate(cycle_nodes):
        if cycle_node is None:
            continue
        else:
            is_active = active_startnodes[ind]
            path = get_path(cycle_node, is_active, graph, ind)
            for edge in iter_edge_in_path(path):
                if is_active:
                    graph.del_edge(edge)
                    is_active = not (is_active)
                else:
                    graph.add_edge((edge[1], edge[0]))
                    is_active = not (is_active)
    return graph


def del_marks(graph, start_nodes):
    for node in range(graph.node_number):
        del_outarc_marks(graph, node)
    return graph


def del_outarc_marks(graph, node):
    node_pointer = graph.nodes[node]
    node_pointer.active_marked = {}  # dictionary out_arrows as key, number of Schlaufe as Value
    node_pointer.passive_marked = {}




import copy
import numpy as np
import unittest

from ugd.help_function.graph_creation import graph_to_adj_m, adj_m_to_graph
from ugd.help_function.util import get_path
from ugd.schlaufen_construction.di_schlaufen_construction_util import mark_edge
from test.test_resources.graphs_two_restriction_sets import graph1, graph1_adj_m


class HelpFunctionsAndUtil(unittest.TestCase):



    def test_adj_m_to_graph(self):
        adj_m = graph1_adj_m
        graph = adj_m_to_graph(adj_m, is_directed=True)
        adj_m_2 = graph_to_adj_m(graph)
        assert np.array_equal(adj_m_2 , adj_m)


    def test_graph_to_adj_m(self):
        graph = graph1
        adj_m = graph1_adj_m
        graph_copy = copy.deepcopy(graph)
        adj_m_verg = graph_to_adj_m(graph_copy)
        assert np.array_equal(adj_m_verg, adj_m)


    def test_get_path(self):
        pathgraph = copy.deepcopy(graph1)
        mark_edge(pathgraph, 3, 2, True, 1)
        mark_edge(pathgraph, 2, 1, False, 1)
        mark_edge(pathgraph, 1, 0, True, 1)
        mark_edge(pathgraph, 0, 3, False, 1)
        active_start = True

        path = get_path(3, active_start, pathgraph, 1)
        assert path == [3,2,1,0,3]


from ugd.help_function.util import rand_element_of_set

''' Random draw functions:'''


def random_draw(graph, working_node, is_active):
    # draw out of feasible out node set
    out_node_set = feasible_out_nodes(graph, working_node, is_active)
    if len(out_node_set) is 0:
        return False, None
    else:
        return True, rand_element_of_set(out_node_set)


def feasible_out_nodes(graph, working_node, is_active):
    # construction of feasible out node set
    out_node_set = set([])
    if is_active:
        for out_node in graph.nodes[working_node].outnodes:
            if is_feasible_out_node(graph, working_node, out_node, is_active):
                out_node_set.add(out_node)
    else:
        for out_node in graph.nodes[working_node].passive_outnodes:
            if is_feasible_out_node(graph, working_node, out_node, is_active):
                out_node_set.add(out_node)
    return out_node_set


def is_feasible_out_node(graph, working_node, out_node, is_active):
    working_node_p = graph.nodes[working_node]
    if is_active:
        if out_node in working_node_p.outnodes and not (out_node in working_node_p.active_marked):
            return True
        else:
            return False
    else:
        if out_node in working_node_p.passive_outnodes and not (out_node in working_node_p.passive_marked) \
                and not (graph.nodes[out_node].outnodes == set([])):
            return True
        else:
            return False


''' mark functions:'''


def mark_node(graph, working_node, is_active):
    if is_active:
        graph.nodes[working_node].active_visited = True
    else:
        graph.nodes[working_node].passive_visited = True


def mark_edge(graph, working_node, out_node, is_active, schlaufen_number):
    if is_active:
        graph.nodes[working_node].active_marked[out_node] = schlaufen_number
    else:
        graph.nodes[working_node].passive_marked[out_node] = schlaufen_number


''' Control functions'''


def cycle_found(graph, out_node, is_active):
    if not is_active:  # consider that this is the new node, which is found in an is_active step but left in an
        # not is_active  step
        if graph.nodes[out_node].active_visited == True:
            return True
        else:
            return False
    else:
        if graph.nodes[out_node].passive_visited == True:
            return True
        else:
            return False



'''
Controlls sequence of all the steps used for a call of the graph_hyp_test function
'''
from ugd.help_function.graph_creation import generate_graph, graph_to_adj_m
from ugd.high_level_interface.output_processing import postprocess
from ugd.high_level_interface.time_mixing_evaluation import evaluate_mixing_time
from ugd.high_level_interface.validation import validate_input, parse_input
from ugd.markov_walk.markov_walk import markov_walk


def hyp_test(adj_m, var_dict, stat_f, test_variable, controlls, mixing_time, anz_sim, show_polt, is_directed, fast_mixing_time_evaluation):
    ''' contains the actual flow of the procedure,'''

    '1) validation and input substitution'
    validate_input(adj_m, anz_sim, mixing_time, var_dict, stat_f, controlls, test_variable, is_directed)
    stat_f, nodesetpartition = parse_input(stat_f, test_variable, controlls, var_dict, adj_m)

    '2) parse graph, into shape for simulation'
    graph = generate_graph(adj_m, nodesetpartition, is_directed)

    '3) determine mixing time, and mixing "amount" '
    mixing_time, edges_changed_per_draw = evaluate_mixing_time(graph, mixing_time, anz_sim, fast_mixing_time_evaluation)

    '4) generating the random graphs under the constraints'
    graphs, stats_list = gen_draws(graph, var_dict=var_dict, stat_f=stat_f, anz_sim=anz_sim, mixing_time=mixing_time)

    '5) post processing, estimating the distribution, plotting'
    plt, info_dict = postprocess(adj_m_original=adj_m, stats_list=stats_list, var_dict=var_dict, stat_f=stat_f,
                                 show_polt=show_polt, test_variable=test_variable,
                                 edges_changed_per_draw=edges_changed_per_draw)

    return {'graph_list': graphs, 'stat_list': stats_list, 'plot': plt, 'info_dict': info_dict}


def gen_draws(graph, mixing_time, anz_sim, var_dict, stat_f):
    'iteratively generating anz_sim random sample graphs'
    graph_list = []
    stat_list = []
    for i in range(anz_sim):
        # a random draw
        graph = markov_walk(graph, mixing_time)
        graph_list.append(graph_to_adj_m(graph))
        adj_m_generated = graph_to_adj_m(graph)
        stat_list.append(stat_f(adj_m_generated, var_dict))

    return graph_list, stat_list



'''User Interface with documentation '''

from ugd.high_level_interface.draw_simulations import hyp_test


def digraph_hyp_test(adj_m, var_dict=None, stat_f=None, test_variable=None, controls=None, mixing_time=None,
                     anz_sim=1000, show_polt=False, fast_mixing_time_evaluation = False):
    '''
    PURPUSE: generation of random Digraphs which a given degree sequence and "crossarrow" restrictions, in order to estimate
             the distribution of a test statistic under the 0 Hypotheses. Further explanation on methods is given in
             paper : Testing Strategic Interaction in Networks

    INPUT:
    :param adj_m:         A numpy array containing 0 and 1s as elements, representing ajdancy matrix of the Digraph
    :param var_dict:      A dictionary with the integer 1..n as primary key, where n is the number of nodes,
                          the Values are dictionaries as well with the variable names as key and the variable value as
                          value the values have to be numbers or strings
    :param stat_f:        A function which maps the adj_m and var_dict to a number "The statistic of interest".
    :param test_variable: Alternative to stat_f, creating a statistic which counts arf form a node subsets into another:
                          a triple with first element variable name, second the value of the variable for the set where
                          the arcs leave and third the value of the subset where the arrow go to
    :param controls:      List of variable names of controls
    :param mixing_time:   Number of runs (steps in the markov graph) before a the graph is considered random
    :param anz_sim:       Number of simulations
    :param show_polt:     Boolean whether a plot is desired
    :return:
    graph_list:           List of random adjacency matrices with the given degree sequence and the cross arrows
    stats_list:           List of the statistics stat_f evaluated for the random graphs

    USAGE:
    import numpy
    adj_m = numpy.zeros((4,4))
    adj_m[0,1] = 1
    adj_m[2,3] = 1
    var_dict ={
        0: {'gender': 'm'},
        1: {'gender': 'm'},
        2: {'gender': 'f'},
        3: {'gender': 'f'},
    }
    graphs, stats_list = digraph_hyp_test(adj_m=adj_m, var_dict = var_dict, test_variable= ('gender','m','f'),mixing_time=10000, anz_sim=100, show_polt=True)

    AUTORS:
    Andrin Pelican, andrin.pelican@unisg.ch

    '''

    return hyp_test(adj_m, var_dict, stat_f, test_variable, controls, mixing_time, anz_sim, show_polt,
                    is_directed=True,  fast_mixing_time_evaluation = fast_mixing_time_evaluation)


def graph_hyp_test(adj_m, var_dict=None, stat_f=None, test_variable=None, controls=None, mixing_time=None,
                   anz_sim=1000, show_polt=False,  fast_mixing_time_evaluation = False):
    '''
    PURPUSE: generation of random Graphs with a given degree sequence and restrictions induced by the node-attributes
             in order to estimate the distribution of a test-statistic under the 0 Hypotheses.Further explanation on
             methods is given in paper:
             Uniform Sampling of Graphs with Fixed Degree Sequence under Partition Constraints

    INPUT:
    :param adj_m:         A numpy array containing 0 and 1s as elements, representing adjacency matrix of the graph
    :param var_dict:      A dictionary with the integers 1..n as primary key (representing the n nodes).
                          The values are dictionaries containing the Variable name as keys and the values can either be
                          numbers or strings.
    :param stat_f:        A function which maps the adj_m and var_dict to a number "the statistic of interest".
    :param test_variable: Alternative to stat_f, creating a statistic which counts the arrows form a node-subset into
                          another. It is a triple with first element variable name, second the value of the variable for
                          the set where the arrows leave and third the value of the subset where the arrow go to.
    :param controls:      List of variable names, the number of arrows crossing the groups induced by the controls is
                          constant in all the simulation.
    :param mixing_time:   Number of runs (steps in the markov graph) before a the graph is considered random
    :param anz_sim:       Number of simulations
    :param show_polt:     Boolean whether a plot is desired

    OUTPUT:
    :return:
    graph_list:           List of random adjacency matrices with the given degree sequence and arrows between the controls
    stats_list:           List of the statistics stat_f evaluated for the random graphs

    USAGE:
    import numpy
    adj_m = numpy.zeros((4,4))
    adj_m[0,1] = 1
    adj_m[1,0] = 1
    adj_m[3,2] = 1
    adj_m[2,3] = 1
    var_dict ={
        0: {'gender': 'm'},
        1: {'gender': 'm'},
        2: {'gender': 'f'},
        3: {'gender': 'f'},
    }
    graphs, stats_list = graph_hyp_test(adj_m=adj_m, var_dict = var_dict, test_variable= ('gender','m','f'),mixing_time=10000, anz_sim=100, show_polt=True)

    AUTORS:
    Andrin Pelican, andrin.pelican@bluewin.ch

    '''

    return hyp_test(adj_m, var_dict, stat_f, test_variable, controls, mixing_time, anz_sim, show_polt,
                    is_directed=False,  fast_mixing_time_evaluation = fast_mixing_time_evaluation)



from ugd import digraph_hyp_test
from ugd import graph_hyp_test

import numpy as np
import matplotlib.pyplot as plt
adj_m = np.zeros((4,4))
adj_m[0,1] = 1
adj_m[2,3] = 1
var_dict ={
    0: {'gender': 'm'},
    1: {'gender': 'm'},
    2: {'gender': 'f'},
    3: {'gender': 'f'},
}

adj_m_2 = np.zeros((6, 6))
adj_m_2[0, 1] = 1
adj_m_2[1, 0] = 1
adj_m_2[1, 2] = 1
adj_m_2[2, 1] = 1
adj_m_2[2, 0] = 1
adj_m_2[0, 2] = 1
adj_m_2[3, 4] = 1
adj_m_2[4, 3] = 1
adj_m_2[4, 5] = 1
adj_m_2[5, 4] = 1
adj_m_2[5, 3] = 1
adj_m_2[3, 5] = 1

var_dict_2 = {
    0: {'gender': 'm'},
    1: {'gender': 'm'},
    2: {'gender': 'f'},
    3: {'gender': 'f'},
    4: {'gender': 'f'},
    5: {'gender': 'f'},
}

out_dict = digraph_hyp_test(adj_m=adj_m, var_dict = var_dict, test_variable=('gender','m','f'), mixing_time=1000, anz_sim=2, show_polt=False)

ax = out_dict['plot']
plt.show()

#out_dict = digraph_hyp_test(adj_m=adj_m, var_dict = var_dict,  mixing_time=1000, anz_sim=2, show_polt=True)

#graphs, stats_list = graph_hyp_test(adj_m=adj_m_2, var_dict = var_dict_2,  mixing_time=1000, anz_sim=100, show_polt=True)





import copy
from ugd.schlaufen_construction.di_schlaufen_construction_util import feasible_out_nodes
from test.test_resources.graphs_two_restriction_sets import graph1
import unittest

testdata = [
    (graph1, 0, False, set([3]) ),
    (graph1, 1, True, set([0]) ),
    (graph1, 0, True, set([]) ),
]

class testFeasibleOutnodes(unittest.TestCase):

    def test_feasable_outnodes(self):

        for graph, node, is_aktive, outnodes in testdata:
            graph_copy = copy.deepcopy(graph)
            outnodes_computed = feasible_out_nodes(graph_copy, node, is_aktive)
            assert outnodes_computed == outnodes



import numpy as np
import random
from ugd.model.digraph import RstDiGraph

import inspect

'''Random function'''


def rand_choise(q):
    ''' returns true with probability q '''
    event = np.random.binomial(1, q)
    return event == 1


def rand_element_of_set(set):
    return random.sample(set, 1)[0]


''' path algorithms, for construction marking and unmarking'''


def get_path(startnode, active_start, graph, pathnumber):
    # This function is only allowed be used on schlaufen of type 1
    path = []
    working_node = startnode
    path_continue = True
    is_aktive = active_start

    while path_continue:
        path.append(working_node)
        path_continue, working_node = get_next_node(graph, working_node, is_aktive, pathnumber)
        is_aktive = not (is_aktive)

    del_nodes_mark(graph)
    return path


def get_next_node(graph, working_node, is_aktive, pathnumber):
    if is_aktive:
        out_dict = graph.nodes[working_node].active_marked
        for key, value in out_dict.items():
            if value == pathnumber:
                if graph.nodes[working_node].active_visited:
                    return False, key
                else:
                    graph.nodes[working_node].active_visited = True
                    return True, key

    else:
        out_dict = graph.nodes[working_node].passive_marked
        for key, value in out_dict.items():
            if value == pathnumber:
                if graph.nodes[working_node].passive_visited:
                    return False, key
                else:
                    graph.nodes[working_node].passive_visited = True
                    return True, key
    return False, None


def del_nodes_mark(graph, startnode=None):
    for node in range(graph.node_number):
        del_node_mark(graph, node)


def del_node_mark(graph, node):
    graph.nodes[node].active_visited = False
    graph.nodes[node].passive_visited = False


def form_to_set_index(graph, edge):
    for ind, set in enumerate(graph.restriction_set_list):
        if edge[0] in set:
            from_ind = ind
        if edge[1] in set:
            to_ind = ind
    return from_ind, to_ind


def all_edges(graph):
    edges = []
    for from_node in range(graph.node_number):
        for to_node in graph.nodes[from_node].outnodes:
            edges.append((from_node, to_node))
    return list(set(edges))


''' Help function for debugging and controll'''


def check_symmetric(a, tol=1e-8):
    return np.allclose(a, a.T, atol=tol)


def isdebugging():
    for frame in inspect.stack():
        if frame[1].endswith("pydevd.py"):
            return True
    return False



import numpy as np
import pytest
import unittest



from ugd import  digraph_hyp_test
from test.test_resources.graphs_two_restriction_sets import graph1_adj_m, graph2_adj_m, var_dict2

var_dict1 = {
    0: {'gender': 'm', 'age': 50},
    1: {'gender': 'm', 'age': 50},
    2: {'gender': 'f', 'age': 51},
    3: {'gender': 'f', 'age': 51},
}

class TestIntegrationCrossingMatrix(unittest.TestCase):

    def test_graph_no_rstrc(self):
        output_dict= digraph_hyp_test(adj_m=graph1_adj_m, var_dict=var_dict1, test_variable=('gender', 'm', 'f'),
                                              anz_sim=1000, mixing_time=100 , show_polt=False)
        stats_list = output_dict['stat_list']
        mue = np.mean(stats_list)
        assert mue > 0.45 and mue < 5.5


    def teste_graph_with_rstrc(self):
        with pytest.raises(ValueError):
            _ = digraph_hyp_test(adj_m=graph1_adj_m, var_dict=var_dict1, test_variable=('gender', 'm', 'f'),
                                 anz_sim=1000, controls=['age'])


    def test_graph2_with_rstrc(self):
        output_dict = digraph_hyp_test(adj_m=graph2_adj_m, var_dict=var_dict2, test_variable=('gender', 'm', 'f'),
                                              anz_sim=1000)
        stats_list = output_dict['stat_list']
        mue = np.mean(stats_list)
        assert mue > 0.33333-0.05 and mue < 0.33333+0.05

    def test_graph2_no_rstrc(self):
         output_dict = digraph_hyp_test(adj_m=graph2_adj_m, var_dict=var_dict2, test_variable=('gender', 'm', 'f'),
                                        anz_sim=1000, controls=['age'])
         stats_list = output_dict['stat_list']
         mue = np.mean(stats_list)
         assert mue > 0.45 and mue < 5.5






"""

In this function the edge probabilities under the 0-Model are estimated. These probabilities are used to calculate
the optimal test statistic.

let n be the number of nodes

Input:
    adjacency matrix: (n, n) np array
    node_dict: dictionary with same structure as in ugd
    controls: list of controls like in ugd

Output:
    Depreciated update comments

Procedure:

    1) parse the graph and node_dict into the matrix input shape for the logit model

    2) estimate the logic model

    3) predict edge probabilities

    4) reshape it into the matrix form of edge probabilities

"""
import logging
logging.getLogger().setLevel(logging.INFO)

import numpy as np
import statsmodels.api as sm
from ugd.locally_most_powerful_stat.util import create_lcl_mst_pwf_statistic_from_thresholds
from ugd.high_level_interface.construct_node_partition import constr_partition


def make_lcl_mst_pwf_stat(adj_m, var_dict=None, controls=None, edge_util_function = None):

    n = adj_m.shape[0]

    # 1) parse the graph and node_dict into the matrix input shape for the logit model
    x,y, edges_removed_due_to_sparsity = create_logit_varaibles(adj_m, var_dict, controls)

    # 2) estimate the logit probabilities
    logit_mod = sm.Logit(endog=y, exog=x)
    logit_res = logit_mod.fit(disp=True, maxiter=20)

    # in order to see whether the minimum is found and whether it is global compare the gradient, with the eigenvalues of the hessian
    logging.info(" The maximal gradient entry in found point is: "+ str(np.max(np.abs(logit_mod.score(logit_res.params)))))
    hessian = - logit_mod.hessian(logit_res.params)
    eigenvalues = np.linalg.eigvalsh(hessian)
    logging.info("Eigenvalues of Hessian: ")
    logging.info("maximal: "+ str(max(eigenvalues)) + "  minimal:   " + str(min(eigenvalues)))

    thresholds = np.dot(x,logit_res.params)
    threholds_m = logit_probs_to_matrix(thresholds,n, edges_removed_due_to_sparsity)

    # creation of the statistic
    optimal_triad_stat = create_lcl_mst_pwf_statistic_from_thresholds(threholds_m, n, edge_util_function = edge_util_function)
    return optimal_triad_stat




def create_logit_varaibles(adj_m, var_dict, controls):
    " bring the matrix shaped node varibles into a table form so that a logit package can be used to estimate"

    n = adj_m.shape[0]
    if not(var_dict is None or controls is None):
        node_set_partition = constr_partition(var_dict=var_dict, controls=controls, adj_m=adj_m)
        sort_partitition(node_set_partition )
        x = np.zeros((n * (n - 1), 2 * n + number_of_possible_crossings(node_set_partition)))
    else:
        x = np.zeros((n * (n - 1), 2 * n ))
    y = np.zeros((n * (n - 1), 1))

    k = 0
    for i in range(n):
        for j in range(n):
            if (j==i):
                continue
            else:
                # target
                y[k, 0] = adj_m[i, j]

                # affinity
                x[k, i] = 1
                x[k, n + j] = 1 # in edges are n values sifted in matrix

                # The indexes for crossing groups, all crossing from one group to another is enumerated
                if not (var_dict is None or controls is None):
                    ind = ind_of_crossing(i, j, node_set_partition)
                    if not(ind is None): # only make entry if nodes are from distinct groups
                        x[k,2*n+ind] = 1
                k+=1

    x,y, edges_removed_due_to_sparsity = remove_sparse_edges(x,y)

    """
    Estimating the node affinity parameters is not possible because of mulitcoliniarity. Increasing
    the in affinity parameters of all nodes by the same amount as decreasing the out affinity of all parameters
    results in the same likelihood. Therefore we take out the inaffinity of the first node.
    """
    x = x[:, 1:]  # the first column is omitted due to multicoliniarity




    """
    Estimating the affinity from and to every group is not possible, 
    because of mulitcoliniarity. By adjusting the out and in affinity of all edges in one gorup
    so that within one group the affinity is the same and to the next node group edges are formed 
    more likely. (the backward edges form the other gorup can be corrected for with the group parameter
    form the other direction) 
    
    Therefore we correct for this fact by for some some parition adjancy matrix edges have
    only coeficients in one direction, such that these parition adjancy matrix edges span
    a tree over the partition graph ( therefore len(node_set_partition)-1 )
    
    due to the colums removed because of sparsity, we dont know whether the last len(node_set_partition)-1 group 
    coefficients span a tree, therefore we check iteratively whether the column is coliniar
    """
    if not(var_dict is None or controls is None):
        # due to sparsity we dont know in advance which columns we have to remove
        for i in reversed(range(x.shape[1])):
            x_reduced = np.delete(x, i, axis=1)
            if np.linalg.matrix_rank(x)==np.linalg.matrix_rank(x_reduced): # check if the colum i is collnear
                x = x_reduced

            if np.linalg.matrix_rank(x) == x.shape[1]:
                break


    # verifying whether there is multicoliniarity. This should not be the case
    if np.linalg.matrix_rank(x) < x.shape[1]:
        print("ATTENTION: the null space of x is not 0, there is multicoliniarity")

    return x,y, edges_removed_due_to_sparsity


def logit_probs_to_matrix(y, n, edges_removed_due_to_sparsity):
    """ shapes node specific values into the matrix which corresponds to the original adjacency matrix
    The two indexes original and reduced is due to the sparse edges"""

    adj_m = np.zeros([n, n])
    ind_y = 0
    ind_original = 0
    for i in range(n):
        for j in range(n):
            if (j == i):
                continue
            elif (ind_original in set(edges_removed_due_to_sparsity)):
                adj_m[i, j] = - 20
                ind_original += 1
            else:
                # target
                adj_m[i, j] = y[ind_y]
                ind_y += 1
                ind_original += 1
    return adj_m

def number_of_possible_crossings(node_set_partition):
    "for each group in partition there is one crossing to each other group "
    return len(node_set_partition)*(len(node_set_partition)-1)


def ind_of_crossing(i,j,node_set_partition):
    "enumerating all except the diagonal elements in the matrix, same principle like in create_logit_varaibles "
    k = 0
    for ind1, set1 in enumerate(node_set_partition):
        for ind2, set2 in enumerate(node_set_partition):
            if (ind1 == ind2):
                continue
            if i in set1 and j in set2:
                return k
            k+=1
    return None


def remove_sparse_edges(x,y):
    """
        In sparse digraphs it often happens that an node has indegree of zero or an outdegree of zero
        or there are no edges between from one nodegroup to another.
        In this case we do not consider these "edges" because the cooeffictints of the node (or nodepartiton)
        would optimize the likeleyhood ad inifinty.

        We therefore exclude the edge and the parameter form the estimation.
        For the optimal test staistic the trasholds have to be set for each potential edge.
        The trash holds of these potential edges are set to minus infinity, so that they dont contribute
        to the teststatistic. All simulated graphs will not have an edge realisation at these potential edges
        due to the degree and the nodegrouprestrictions.
        """


    sparse_columns = []
    sparse_edges = []
    # loop over columns in order to find sparse in/out nodes or groups
    for i in range(x.shape[1]):

        column = x[:, i]
        is_sparse = True
        for j in range(x.shape[0]):
            if column[j] == 1 and y[j] == 1:
                is_sparse = False

        if is_sparse:
            sparse_columns.append(i)
            # find the corresponding edges
            for j in range(x.shape[0]):
                if column[j] == 1:
                    sparse_edges.append(j)

    sparse_edges = list(set(sparse_edges)) # remove duplicates

    # remove edges
    y = np.delete(y, sparse_edges, 0)
    x = np.delete(x, sparse_edges, 0)

    # remove the corresponding varibales
    x = np.delete(x, sparse_columns,1)

    logging.info("There are "+str(sparse_edges.__len__()) + " not considered in estimation, becaues they belong to a spares "
                                                     "node (one with no in or edge realisation at all or due to a "
                                                     "sparse group constraint")

    return x,y,sparse_edges



def sort_partitition(node_set_partition):
     """The set partition has naturally no order,
     For the sparse graph, the order can be important when deleting crossgroup varibales in order to avoid
     multicoliniarity.  This function induces a fixed order (at least in the cases where the elsments sum is unique for
     each element)"""
     node_set_partition.sort(reverse=False,  key=lambda x: sum(x))


from ugd.model.dinode import DiNode

'''
The DiGraph contains the information on the degrees and nodes. 
The RstDiGraph contains in addition information on the restrictions (here, the number of arrows between node groups, 
or node partitions) 
'''


class DiGraph():
    def __init__(self, indegree_serie, outdegree_serie):
        self.is_directed = True
        self.indegree_serie = indegree_serie
        self.outdegree_serie = outdegree_serie
        self.node_number = outdegree_serie.__len__()
        self.nodes = []
        for i in range(indegree_serie.__len__()):
            self.nodes.append(DiNode(i, indegree_serie[i], outdegree_serie[i], self.node_number))

    def add_edge(self, edge):
        from_id, to_id = edge
        from_node = self.nodes[from_id]
        from_node.add_outarrow(to_id)
        to_node = self.nodes[to_id]
        to_node.del_passive_outarrow(from_id)

    def del_edge(self, edge):
        from_id, to_id = edge
        from_node = self.nodes[from_id]
        from_node.del_outarrow(to_id)
        to_node = self.nodes[to_id]
        to_node.add_passive_outarrow(from_id)


class RstDiGraph(DiGraph):
    # Graph extension with restriction:
    def __init__(self, indegree_serie, outdegree_serie, restriction_set_list, crossing_np_array):
        "fixed parameter"
        self.restriction_set_list = restriction_set_list  # list of the restriction subsets, disjunct nodepartition
        self.crossing_matrix = crossing_np_array

        "parameters changesed in swich search"
        # stard node nodes for schlaufen
        self.startnodes = []
        # if schlaufe constains cycle, cycle start node, otherwise 0
        self.cycle_start_nodes = []

        self.violation_matrix = None
        # build and inserted after random draw from
        super(RstDiGraph, self).__init__(indegree_serie, outdegree_serie)



class DiNode():
    '''
    The DiNode class is specifically designed for altering path search.

    -  Active and passive out-nodes, for easily iteratively find new nodes in path.
    -  Marks for easy look up arrows already visited (earlier Schlaufen) / nodes in path-creation already visited
    -  The arrow-marks have the Schlaufen number saved, such that later a reconstruction of the Schlaufen found is possible
    '''

    def __init__(self, id, max_indegree, max_outdegree, totalnodes):
        self.id = id
        self.indegree = max_indegree
        self.outdegree = max_outdegree

        #
        self.passive_outnodes = set(range(totalnodes))
        self.passive_outnodes.remove(id)
        self.passive_outnodes_removed_selfloop_crossarrow = 1  # for graph correctness check
        self.outnodes = set()

        # marks
        self.active_visited = False
        self.passive_visited = False

        # marked edges
        self.active_marked = {}  # dictionary outarrows as key, number of schlaufe as Value
        self.passive_marked = {}

    def add_outarrow(self, out_node):
        self.outnodes.add(out_node)

    def add_passive_outarrow(self, out_node):
        self.passive_outnodes.add(out_node)

    def del_outarrow(self, out_node):
        if not (out_node in self.outnodes):
            raise ValueError('outnode subtraction error')
        else:
            self.outnodes.remove(out_node)

    def del_passive_outarrow(self, out_node):
        if not (out_node in self.passive_outnodes):
            raise ValueError('outnode subtraction error')
        else:
            self.passive_outnodes.remove(out_node)



from test.test_resources.nyakatoke_data.nyakatoke_network import di_adj_m, var_dict
import numpy as np
import unittest

from ugd import digraph_hyp_test
from ugd.help_function.graph_creation import adj_to_in_out_sequences, generate_graph, get_crossing_matrix
from ugd.high_level_interface.construct_node_partition import constr_partition


anz_sim = 3


class TestIntegrationCrossingMatrix(unittest.TestCase):

    # wealth and religion is unrelated nearly
    def test_assert_crossingmatrix(self):
        di_adj_m1 =di_adj_m
        out_dict = digraph_hyp_test(adj_m=di_adj_m,var_dict=var_dict,test_variable=('wealth','rich','poor'),anz_sim=anz_sim, show_polt=False, controls=['wealth'])
        partition = constr_partition(controls =['wealth'], var_dict= var_dict, adj_m=di_adj_m)

        di_adj_m2 = out_dict['graph_list'][anz_sim - 1]
        _, ins1, outs1 = adj_to_in_out_sequences(di_adj_m1)
        _, ins2, outs2 = adj_to_in_out_sequences(di_adj_m2)

        crossing_matrix1 = generate_graph(di_adj_m1, partition, is_directed=True).crossing_matrix
        crossing_matrix2 = generate_graph(di_adj_m2, partition, is_directed=True).crossing_matrix
        np.testing.assert_equal(crossing_matrix1, crossing_matrix2)
        np.testing.assert_equal(ins1, ins2)




'''
Control functions:
Performing checks on the Graph, to see whether implicit restrictions (double arrow, degree  ect) are fulfilled.
Checks whether there are no contradicting graph data saved (in case of redundancy)
'''

import numpy as np
from ugd.help_function.util import all_edges, form_to_set_index

'DIGRAPH:'


def graph_correct(graph):
    restriction(graph)
    graph_nodes_correct(graph)
    return True


def full_graph_correct(graph):
    restriction(graph)
    graph_nodes_correct(graph)
    graph_is_full(graph)
    matrix_consistent(graph)
    return True


def graph_nodes_correct(graph):
    node_id_in_place(graph)
    for node in graph.nodes:
        node_correct(node, graph.node_number)
    return True


def restriction(rstrGraph):
    # checks whether partition input is correct (disjoint partition of integers {0,..,n-1}
    set_list = rstrGraph.restriction_set_list
    total_Set = set()
    total = 0
    for sub_set in set_list:
        total_Set = total_Set.union(sub_set)
        total += sub_set.__len__()
    if not (total == total_Set.__len__()):
        raise ValueError('restriction sets are not disjoint')
    if not (total_Set == set(range(rstrGraph.outdegree_serie.__len__()))):
        raise ValueError('nodes in set are not the intergers {0,..n-1}')


def graph_is_full(graph):
    # checks whether the digraph degrees are consistent with the degree sequence
    for i, node in enumerate(graph.nodes):
        if not (graph.indegree_serie[i] == node.indegree):
            raise ValueError('indegree property not fulfielled')
        if not (graph.outdegree_serie[i] == node.outdegree):
            raise ValueError('indegree property not fulfielled')


def matrix_consistent(rstGraph):
    # controls whether the graph fulfills the partition restriction
    n = rstGraph.restriction_set_list.__len__()
    controll_matrix = np.zeros((n, n))
    for edge in all_edges(rstGraph):
        from_ind, to_ind = form_to_set_index(rstGraph, edge)
        if not (from_ind == to_ind):
            controll_matrix[from_ind, to_ind] += 1
    if not (controll_matrix == rstGraph.crossing_matrix).all():
        raise ValueError('Graph has the setrestricions violated')


'NODE:'


def node_id_in_place(graph):
    for i, node in enumerate(graph.nodes):
        if not (i == node.id):
            raise ValueError('id and node position does not match')


def node_correct(node, n):
    if not (node.outdegree == node.outnodes.__len__()):
        raise ValueError('degree violated !!')
    if not (node.indegree == n - node.passive_outnodes.__len__() - node.passive_outnodes_removed_selfloop_crossarrow):
        raise ValueError('degree violated !!')

    for outnode in node.outnodes:
        if outnode == node.id:
            raise ValueError('self loop')



import numpy as np
from ugd.model.graph import RstGraph
from ugd.schlaufen_construction.schlaufen_construction import add_plain_random_schlaufe

import unittest

'''
There are only 2 graph with this restrictions for switching it the path has to go over all edges

The test idea is that if the path starts at 1 and goes to 0 then over {3,4} back to 0, then with probability 1/2
a schlaufe of type 3 is build, or the path is continued. We test whether the probability is indeed 1/2.

'''

def new_graph():
    graph = RstGraph(degree_serie=[2, 1, 1, 1, 1], restriction_set_list=[set([0]), set([1, 2]), set([3, 4])],
                     crossing_np_array=np.array(([0, 2, 0], [2, 0, 0], [0, 0, 1])))
    graph.add_edge((0, 1))
    graph.add_edge((0, 2))
    graph.add_edge((3, 4))
    return graph

class SchlaufeType3(unittest.TestCase):

    def test_schlaufe_type_3(self):
        not_case_3 = 0
        case_3 = 0

        for i in range(10000):
            graph = new_graph()

            start_node, cycle_node, active_cycle_node = add_plain_random_schlaufe(graph, schleifen_number=0)
            if start_node== 1:

                # case type 3
                if 2 in graph.nodes[0].active_marked: # the path continued to 2
                    # not case 3 at 0 link
                    not_case_3 += 1
                else:
                    if 0 in graph.nodes[3].passive_marked or  0 in graph.nodes[4].passive_marked: # the path came to 0 but didnt contunue to 2
                        case_3 += 1
        assert abs(not_case_3-case_3)/(case_3+not_case_3)<0.1


import numpy as np
import unittest
from ugd import graph_hyp_test
from test.test_resources.plain_graph_integration import adj_m_1, var_dict_1, adj_m_2, var_dict_2

class TestIntegrationGraph(unittest.TestCase):

    def test_graph_no_rstrc(self):
        output_dict= graph_hyp_test(adj_m=adj_m_1, var_dict=var_dict_1, test_variable=('gender', 'm', 'f'), mixing_time=100, anz_sim=1000, show_polt=False)
        mue = np.mean(output_dict['stat_list'])
        true_mean = 0*1/3+ 2/3 * 2 # in this simple case the true mean can be calculated manually
        assert mue > true_mean - 0.15 and mue < true_mean + 0.15




import matplotlib.pyplot as plt
from ugd.help_function.util_statistic import get_quantile


def postprocess(adj_m_original, stats_list, stat_f, var_dict, test_variable, edges_changed_per_draw, show_polt=False):
    '''
    General output postprocessing and plotting, this part is convenient for experimenting with graph
    However can be easily customized by using the output graph sequence of the algorithm.
    '''
    org_value = stat_f(adj_m_original, var_dict)
    quantile = get_quantile(org_value, stats_list)

    statisitc_name = stat_f.__name__
    if not (test_variable == None):
        statisitc_name = 'crosslinks ' + str(test_variable[0]) + ': ' + str(test_variable[1]) + ' to ' + str(test_variable[2])

    info_dict = {
        'stat_name': statisitc_name,
        'original_value': org_value,
        'quantile': quantile,
    }

    annotate_string = 'Distribution estimate of the test statistic:   ' + statisitc_name + '\n' \
                      + 'The observed value lies in quantile:            ' + str(quantile) + '\n' \
                      + 'Average changed edges per draw:              ' + str(int(edges_changed_per_draw * 100)) + '%'

    min_value = min(stats_list)
    max_value = max(stats_list)
    y = [0] * (int(max_value + 1) - int(min_value))
    x = list(range(int(min_value), int(max_value + 1)))
    for stat_val in stats_list:
        y[int(stat_val) - int(min_value)] += 1 / stats_list.__len__()

    width = 1 / 1.5
    plt.bar(x, y, width, color="blue")
    plt.axvline(x=org_value, color='red')
    plt.title('Distribution')
    plt.ylabel('Probability')
    plt.xlabel('Values')
    plt.gcf().subplots_adjust(bottom=0.25)
    plt.annotate(annotate_string, (0, 0), (0, -35), xycoords='axes fraction', textcoords='offset points', va='top')

    if show_polt:
        plt.show()

    return plt, info_dict



from ugd.schlaufen_construction.schlaufen_construction import add_plain_random_schlaufe
from ugd.schlaufen_construction.di_schlaufen_construction import add_di_random_schlaufe


def add_random_schlaufe(graph, schleifen_number):
    # directs to the right Schlaufen construction function, directed and undirected,
    if graph.is_directed:
        return add_di_random_schlaufe(graph, schleifen_number)
    else:
        return add_plain_random_schlaufe(graph, schleifen_number)



'''

This file implements the Schlaufen Detection algorithm in the paper:

Testing Strategic Interaction in Networks

'''

from ugd.schlaufen_construction.di_schlaufen_construction_util import mark_edge, mark_node, cycle_found, random_draw
from ugd.help_function.util import rand_element_of_set, del_nodes_mark


def add_di_random_schlaufe(graph, schlaufen_number):
    # random draw of initial node (step 1)
    start_node = rand_element_of_set(range(graph.node_number))

    cycle_node = None
    active_cycle_node = None
    working_node = start_node

    is_active = True
    is_schlaufe = False
    while not is_schlaufe:
        mark_node(graph, working_node, is_active)
        found, out_node = random_draw(graph, working_node, is_active)  # step 2 or 4, depending on is_active
        if found:
            mark_edge(graph, working_node, out_node, is_active, schlaufen_number)
            working_node = out_node
            if cycle_found(graph, out_node, is_active): # step 3 or 5
                active_cycle_node = not is_active  # out_node is one step ahead
                cycle_node = out_node
                is_schlaufe = True
            else:
                is_active = not is_active
        else:
            is_schlaufe = True

    del_nodes_mark(graph, start_node)   # but don't delete marked edges
    return start_node, cycle_node, active_cycle_node



'''
Constructs a node partition, according to the variables the user wants to control for: "controls".
The controls are keys in the variable dictionary, the node-groups (elements of the partition) are
the nodes which share the values for all the controls.

This input from for is not a restriction, any partition can be defined via node-variables.
'''


def constr_partition(controls, var_dict, adj_m):
    if controls == None:
        nodesetpartition = [set(range(adj_m.shape[0]))]
    else:
        # fist step: create new variable out of combination of covariats
        hash_set = set([])
        for node in var_dict:
            hash_set.add(my_hash(controls, var_dict, node))

        # construct set list of sets (node parition)
        nodesetpartition = []
        for place, hash in enumerate(hash_set):
            nodesetpartition.append(set([]))
            for node in var_dict:
                comp_has = my_hash(controls, var_dict, node)
                if my_hash(controls, var_dict, node) == hash:
                    nodesetpartition[place].add(node)
    return nodesetpartition


def my_hash(controlls, var_dict, node):
    has_str = ''
    for controll in controlls:
        has_str += str(var_dict[node][controll]) + '_'
    return has_str



from ugd.help_function.util import form_to_set_index
from ugd.help_function.util import all_edges
from ugd.model.digraph import RstDiGraph
from ugd.model.graph import RstGraph

import numpy as np
import copy

'''
This functions generate from the input a graph in of the class defined in model.
That form is convenient and more efficient for the Schlaufen construction
'''


def generate_graph(adj_m, nodesetpartition, is_directed):
    graph = adj_m_to_graph(adj_m, is_directed)
    graph.restriction_set_list = nodesetpartition
    crossing_martix = get_crossing_matrix(graph)
    graph.crossing_matrix = crossing_martix
    graph = postprocess_graph(graph)
    return graph


def get_crossing_matrix(graph):
    crossing_matrix = np.zeros((graph.restriction_set_list.__len__(), graph.restriction_set_list.__len__()))
    edges = all_edges(graph)
    for edge in edges:
        from_ind, to_ind = form_to_set_index(graph, edge)
        if graph.is_directed:
            crossing_matrix[from_ind, to_ind] += 1
        else:
            crossing_matrix[from_ind, to_ind] += 1
            crossing_matrix[to_ind, from_ind] += 1
    return crossing_matrix


def graph_to_adj_m(graph):
    adj_m = np.zeros(shape=(graph.node_number, graph.node_number))
    for ind, node in enumerate(graph.nodes):
        for out_ind in node.outnodes:
            adj_m[ind, out_ind] += 1
    return adj_m


def adj_m_to_graph(adj_m, is_directed):
    if is_directed:
        n, indegrees, outdegrees = adj_to_in_out_sequences(adj_m)
        graph = RstDiGraph(indegree_serie=indegrees, outdegree_serie=outdegrees, restriction_set_list=None,
                           crossing_np_array=None)
        for i in range(n):
            for j in range(n):
                if adj_m[i, j] == 1:
                    graph.add_edge((i, j))
    else:
        n, degrees = adj_to_sequences(adj_m)
        graph = RstGraph(degree_serie=degrees, restriction_set_list=None, crossing_np_array=None)
        for i in range(n):
            for j in range(i, n):
                if adj_m[i, j] == 1:
                    graph.add_edge((i, j))

    return graph


def adj_to_in_out_sequences(adj):
    n = adj.shape[0]
    indegrees = []
    outdegrees = []
    for i in range(n):
        outdegrees.append(sum(adj[i, :]))
        indegrees.append(sum(adj[:, i]))
    return n, indegrees, outdegrees


def adj_to_sequences(adj):
    n = adj.shape[0]
    degrees = []
    for i in range(n):
        degrees.append(sum(adj[i, :]))
    return n, degrees


''' 
The Following function:

Updates the passive_out nodes for each node, taking out the out nodes which are not feasible because the
new node out degree 0, or a switch would imply a creation of an edge which is never in the graph. This makes 
the algorithm more efficient in handling hard constraints.

For a formal proof of correctness see:

Uniform Sampling of Graphs with Fixed Degree Sequence under Partition Constraints, Section 2.4

'''


def postprocess_graph(graph):
    for node_p in graph.nodes:
        node_p = reduce_passive_outnodes(graph, node_p)
    return graph


def reduce_passive_outnodes(graph, node_p):
    for passive_outnode in copy.deepcopy(node_p.passive_outnodes):
        if not_feasable_pas_outnode(graph, node_p.id, passive_outnode):
            node_p.passive_outnodes.remove(passive_outnode)
            node_p.passive_outnodes_removed_selfloop_crossarrow += 1
    return node_p


def not_feasable_pas_outnode(graph, node, passive_outnode):
    # chek outdegree
    if graph.is_directed:
        if graph.nodes[passive_outnode].outdegree == 0:
            return True
    else:
        if graph.nodes[passive_outnode].degree == 0:
            return True

    i, j = form_to_set_index(graph, (node, passive_outnode))
    if graph.crossing_matrix[j, i] == 0:
        return True
    return False



import json
from test.test_util.json_citation_read_in import createGraph
from ugd import digraph_hyp_test



'''loading and preparing'''
inputPaperList = json.load(open('../test/test_resources/output_24_12_17.json'))
adj_m, var_dict = createGraph(inputPaperList, [2018, 2015, 2012, 2009], 'econometric theory' )

'''actual estimation'''
graphs, stats_list = digraph_hyp_test(adj_m=adj_m, var_dict = var_dict, test_variable= ('journal','econometric theory','econometric review'), controls=['timelayer', 'journal'], anz_sim=50, show_polt=True)

pass




import numpy as np

# reduction cyrcle

indegree_serie_1 =[1, 0, 1, 0]
outdegree_serie_1 =[0, 1, 0, 1]
restriction_set_list_1 =[set([0, 3]), set([1, 2])]
crossing_np_array_1 = np.array(([0, 0], [0, 0]))

# aumentation cyrcle

indegree_serie_2=[1,0,1,0]
outdegree_serie_2=[0,1,0,1]
restriction_set_list_2=[set([0,1]),set([2,3])]
crossing_np_array_2 = np.array(([0,1],[1,0]))

# long cycle
indegree_serie_3 =[0,1,0,2,0,2,0,1]
outdegree_serie_3 =[1,0,2,0,2,0,1,0]
restriction_set_list_3=[set([0,1,2,3]),set([4,5,6,7])]
crossing_np_array_3 = np.array(([0,0],[0,0]))




# impossible long
indegree_serie_4 =[2,0,2,1,1,2,0]
outdegree_serie_4=[1,1,0,3,2,0,1]
restriction_set_list_4=[set([0,1,2,3]),set([4,5,6])]
crossing_np_array_4 = np.array(([0,0],[0,0]))


# possible not equlibrated:
# long cycle
indegree_serie_5 =[0,2,1,1,1,1,1,1]
outdegree_serie_5 = [2,0,0,3,0,1,1,1]
restriction_set_list_5=[set([0,1,2,3,4]),set([5,6,7])]
crossing_np_array_5 = np.array(([0,0],[0,0]))




import pandas as pd
import numpy as np


# The Data
try:
    Nyakatoke_ind = pd.read_stata("test/test_resources/nyakatoke_data/Nyakatoke individual.dta")
    Nyakatoke_dyad = pd.read_stata("test/test_resources/nyakatoke_data/Nyakatoke dyadic.dta")
    Nyakatoke_hh = pd.read_stata("test/test_resources/nyakatoke_data/Nyakatoke household.dta")
    Nyakatoke_dir = pd.read_stata("test/test_resources/nyakatoke_data/Nyakatoke directed.dta")
except:
    Nyakatoke_ind = pd.read_stata("test_resources/nyakatoke_data/Nyakatoke individual.dta")
    Nyakatoke_dyad = pd.read_stata("test_resources/nyakatoke_data/Nyakatoke dyadic.dta")
    Nyakatoke_hh = pd.read_stata("test_resources/nyakatoke_data/Nyakatoke household.dta")
    Nyakatoke_dir = pd.read_stata("test_resources/nyakatoke_data/Nyakatoke directed.dta")



'''
Data Preperation:
'''

Nyakatoke_ind['hh1']=Nyakatoke_ind['hhind'].apply(str).str[2:5]      # Create household ID number
Nyakatoke_ind['iid']=Nyakatoke_ind['hhind'].apply(str).str[5:7]      # Create individual ID number

# Find age of household head
Nyakatoke_ind['head_age'] = None
Nyakatoke_ind.loc[Nyakatoke_ind['iid']=='01', 'head_age'] = Nyakatoke_ind['age']

# Find sex of household head
Nyakatoke_ind['head_sex'] = None
Nyakatoke_ind.loc[Nyakatoke_ind['iid']=='01', 'head_sex'] = Nyakatoke_ind['sex']


# Now work with the dyadic dataframe
# Convert hh1 to string and add leading zeros (to matched with hh dataframe created above)
Nyakatoke_dyad.loc[:,'hh1'] = Nyakatoke_dyad['hh1'].apply(str).str.zfill(5)

# remove decimal and digits to its right
Nyakatoke_dyad.loc[:,'hh1'] = Nyakatoke_dyad['hh1'].apply(str).str[0:3]

# Repeat conversions for hh2
Nyakatoke_dyad.loc[:,'hh2'] = Nyakatoke_dyad['hh2'].apply(str).str.zfill(5)
Nyakatoke_dyad.loc[:,'hh2'] = Nyakatoke_dyad['hh2'].apply(str).str[0:3]


# NOTE: Last merge creates two instances of hh1 due to how the merge above was done. Drop the second instance and
#       rename the first
# Nyakatoke_dyad.drop('hh1_y', axis=1, inplace=True)
Nyakatoke_dyad.rename(columns={'hh1_x' : 'hh1'}, inplace=True)

Nyakatoke_dyad.rename(columns={'head_age' : 'head_age2', 'head_sex' : 'head_sex2', 'education' : 'education2'}, inplace=True)

'''
add wealth
'''

# Comola and Fafchamps (2014) wealth formula# Comol
Nyakatoke_dyad['wealth1'] = (300000*Nyakatoke_dyad['land1'] + Nyakatoke_dyad['livestock1'])/100000
Nyakatoke_dyad['wealth2'] = (300000*Nyakatoke_dyad['land2'] + Nyakatoke_dyad['livestock2'])/100000
# Group households by wealth
# Very Poor
set1 = set([row['hh1'] for index, row in Nyakatoke_dyad.iterrows() if (row['wealth1'] < 1.5)])
very_poor = list(set1 )

# Poor
set1 = set([row['hh1'] for index, row in Nyakatoke_dyad.iterrows() if (row['wealth1'] >= 1.5) & (row['wealth1'] < 3)])
poor = list(set1 )

# Middle
set1 = set([row['hh1'] for index, row in Nyakatoke_dyad.iterrows() if (row['wealth1'] >= 3) & (row['wealth1'] < 6)])
middle = list(set1 )

# Rich
set1 = set([row['hh1'] for index, row in Nyakatoke_dyad.iterrows() if (row['wealth1'] >= 6)])
rich = list(set1 )






'''
Create adj _m and variable_dict
'''
var_dict = {}
id_to_ind_dict = {}
for i, id in enumerate(Nyakatoke_dyad['hh1'].unique()):
    id_to_ind_dict[id] = i
    var_dict[i] = {} # initialize vardict


# adjancy matrix:
adj_m = np.zeros(shape=(id_to_ind_dict.__len__(),id_to_ind_dict.__len__()))
for key, row in Nyakatoke_dyad.iterrows():
    if row['links']=='reciprocal link' or row['links']=='unilateral link' :
        if row['hh1'] == '122' or row['hh2'] == '122':
            continue
        from_int = id_to_ind_dict[row['hh1']]
        to_int = id_to_ind_dict[row['hh2']]
        adj_m[from_int,to_int] = 1
        adj_m[to_int,from_int] = 1

# adjancy matrix:
di_adj_m = np.zeros(shape=(id_to_ind_dict.__len__(), id_to_ind_dict.__len__()))
for key, row in Nyakatoke_dyad.iterrows():
    if row['links'] == 'reciprocal link' or row['links'] == 'unilateral link':
        if row['hh1'] == '122' or row['hh2'] == '122':
            continue
        from_int = id_to_ind_dict[row['hh1']]
        to_int = id_to_ind_dict[row['hh2']]
        di_adj_m[from_int, to_int] = 1



# var_dict
for key, row in Nyakatoke_dyad.iterrows():
    index_int = id_to_ind_dict[row['hh1']]
    var_dict[index_int]['clan'] = row['clan1']
    # var_dict[index_int]['education'] = row['education1']
    var_dict[index_int]['religion'] = row['religion1']
    if row['land1']>0.5:
        var_dict[index_int]['land']  = 'has_land'
    else:
        var_dict[index_int]['land'] = 'no_land'
    if row['livestock1']>10000:
        var_dict[index_int]['animals'] = 'many'
    else:
        var_dict[index_int]['animals'] = 'few'


for key in very_poor:
    index_int = id_to_ind_dict[key]
    var_dict[index_int]['wealth_granular'] = 'very_poor'
    var_dict[index_int]['wealth'] = 'poor'

for key in poor:
    index_int = id_to_ind_dict[key]
    var_dict[index_int]['wealth_granular'] = 'poor'
    var_dict[index_int]['wealth'] = 'poor'

for key in middle:
    index_int = id_to_ind_dict[key]
    var_dict[index_int]['wealth_granular'] = 'middle'
    var_dict[index_int]['wealth'] = 'rich'

for key in rich:
    index_int = id_to_ind_dict[key]
    var_dict[index_int]['wealth_granular'] = 'rich'
    var_dict[index_int]['wealth'] = 'rich'





'''
Estimates the time of the whole simulation,
estimates the average number of edges changed, in one time step (in % of totoal edges)

'''
import copy
import logging
import time

from ugd.markov_walk.markov_walk import markov_walk

logging.getLogger().setLevel(logging.INFO)


def evaluate_mixing_time(graph, mixing_time, anz_sim, fast_mixing_time_evaluation):
    # preparation

    edges_changed_numb = 0

    if fast_mixing_time_evaluation:
        runs = 10,
    else:
        runs = 1000

    # counting
    overheattime = 0
    start = time.clock()
    for i in range(runs):
        s1 = time.clock()
        compare_graph = copy.deepcopy(graph)
        e1 = time.clock()

        graph = markov_walk(graph, 1)

        s2 = time.clock()
        edges_changed_numb += edges_changed(graph, compare_graph)
        e2 = time.clock()
        overheattime += e1 - s1 + e2 - s2
        # print('run number:    ' + str(i))

    stop = time.clock()

    # validation
    if not(fast_mixing_time_evaluation) and edges_changed_numb == 0:
        raise ValueError('After ' + str(
            runs) + ' runs no other graph has been found, Either there doesent exist one, or the probability of '
                    'finding one is very small and it is recommended to reconsider the problem or use '
                    'a different estimation method.')
    # completion
    n_edges = number_of_edges(graph)
    if mixing_time == None:
        if edges_changed_numb==0:
            logging.critical("no edges are modified while evaluating mixing time, cannot set a default mixing time. Set mixing time manually.")
        mixing_time = int(10 / edges_changed_numb * runs * n_edges)

    # evaluation

    time_taken = stop - start - overheattime
    time_per_run = time_taken / runs
    time_per_graph_creation = time_per_run * mixing_time;
    time_estimated = anz_sim * time_per_graph_creation
    edges_changed_per_draw = edges_changed_numb * mixing_time / runs / n_edges
    logging.info('The time per simulated graph is:  ' + str(time_per_graph_creation) + '  seconds.')
    logging.info('The total estimation will last approximately for:  ' + str(int(time_estimated)) + '  seconds.')
    logging.info('Approximate edge changes per draw of simulated graph in percent of total edges:  ' + str(
        int(edges_changed_per_draw * 100)) + '%.')
    return mixing_time, edges_changed_per_draw


''' Costume functions '''


def edges_changed(graph, comparegraph):
    # counts the edges which are different between the two
    counter = 0
    for ind in range(graph.node_number):
        counter += set_difference_card(graph.nodes[ind].outnodes, comparegraph.nodes[ind].outnodes)
    return counter


def set_difference_card(set1, set2):
    # returns the number of elements in set 1 which are not in set two
    # for nodes this are the edges changed
    couter = 0
    for element in set1:
        if not (element in set2):
            couter += 1
    return couter


def number_of_edges(graph):
    counter = 0
    for node in graph.nodes:
        counter += len(node.outnodes)
    return counter



from ugd.model.node import Node

'''
The Graph contains the information on the degrees and edges, 
The RstGraph contains in addition information on the restrictions (here, the number of arrows between node groups, or 
node partitions) 
'''


class Graph():

    def __init__(self, degree_serie):
        self.is_directed = False
        self.degree_serie = degree_serie
        self.node_number = degree_serie.__len__()
        self.nodes = []
        for i in range(degree_serie.__len__()):
            self.nodes.append(Node(i, degree_serie[i], self.node_number))

    def add_edge(self, edge):
        from_id, to_id = edge
        from_node = self.nodes[from_id]
        from_node.add_outedge(to_id)
        from_node.del_passive_outedge(to_id)
        to_node = self.nodes[to_id]
        to_node.add_outedge(from_id)
        to_node.del_passive_outedge(from_id)

    def del_edge(self, edge):
        from_id, to_id = edge
        from_node = self.nodes[from_id]
        from_node.del_outedge(to_id)
        from_node.add_passive_outedge(to_id)
        to_node = self.nodes[to_id]
        to_node.del_outedge(from_id)
        to_node.add_passive_outedge(from_id)


class RstGraph(Graph):
    # Graph extension with restriction:
    def __init__(self, degree_serie, restriction_set_list, crossing_np_array):
        "fixed parameter"
        self.restriction_set_list = restriction_set_list  # list of the restriction subsets, disjunct nodepartition
        self.crossing_matrix = crossing_np_array

        "parameters changesed in swich search"
        # start node nodes for Schlaufen
        self.startnodes = []
        # if Schlaufe contains cycle, cycle start node, otherwise 0
        self.cycle_start_nodes = []

        self.violation_matrix = None
        # build and inserted after random draw from
        super(RstGraph, self).__init__(degree_serie)



'''
Utility functions for statistic:
Graph statistic and evaluation of quartiles.
'''

import numpy as np
from ugd.help_function.util import check_symmetric


def get_quantile(true_val, sim_values):
    numb_lower = 0
    for value in sim_values:
        if value < true_val:
            numb_lower += 1
        if value == true_val:
            numb_lower += 0.5
    quant = numb_lower / sim_values.__len__()
    return quant


def get_default_stat(adj_m):
    # reciprocity statistic count number of bidirectional arrows (for directed networks)
    def reciprocity(adj_m, var_dict):
        return np.trace(adj_m.dot(adj_m)) / 2

    # transitivity statistic count number triads for a undirected network
    def transitivity(adj_m, var_dict):
        return np.trace(adj_m.dot(adj_m.dot(adj_m))) / 3 / 2 # /2 due to both direction of the triad

    # decide which one to return:
    if check_symmetric(adj_m):
        return transitivity
    else:
        return reciprocity


def crossarrow_count(adj_m, var_dict, test_border):
    key, from_value, to_value = test_border
    count = 0;
    n = adj_m.shape[0]
    for i in range(n):
        for j in range(n):
            if adj_m[i, j] == 1 and var_dict[i][key] == from_value and var_dict[j][key] == to_value:
                count += 1
    return count



from test.test_resources.nyakatoke_data.nyakatoke_network import adj_m, var_dict
import unittest

import numpy as np


from ugd import  digraph_hyp_test, graph_hyp_test
from ugd.help_function.graph_creation import adj_to_in_out_sequences, generate_graph, get_crossing_matrix
from ugd.high_level_interface.construct_node_partition import constr_partition

class TestIntegrationNayatoke(unittest.TestCase):

    def test_assert_crossingmatrix(self):
        anz_sim = 3
        # wealth and religion is unrelated nearliy
        adj_m1 = adj_m
        out_dict = graph_hyp_test(adj_m=adj_m, var_dict=var_dict, test_variable=('wealth', 'rich', 'poor'), anz_sim=anz_sim,
                                  show_polt=False, controls=['wealth'])
        partition = constr_partition(controls=['wealth'], var_dict=var_dict, adj_m=adj_m)

        adj_m2 = out_dict['graph_list'][anz_sim - 1]
        _, ins1, outs1 = adj_to_in_out_sequences(adj_m1)
        _, ins2, outs2 = adj_to_in_out_sequences(adj_m2)

        crossing_matrix1 = generate_graph(adj_m1, partition, is_directed=False).crossing_matrix
        crossing_matrix2 = generate_graph(adj_m2, partition, is_directed=False).crossing_matrix

        assert np.array_equal(crossing_matrix1,crossing_matrix2) and np.array_equal(ins1,ins2)




import numpy as np
import copy
from ugd.help_function.graph_creation import graph_to_adj_m
from ugd.help_function.control_functions_digraph import graph_correct
from ugd.model.digraph import RstDiGraph
from ugd.schlaufen_construction.di_schlaufen_construction_util import mark_edge

# reduction cyrcle
graph1 = RstDiGraph(indegree_serie=[1, 0, 1, 0], outdegree_serie=[0, 1, 0, 1], restriction_set_list=[set([0, 3]), set([1, 2])], crossing_np_array = np.array(([0, 0], [0, 0])))
graph1.add_edge((1, 0))
graph1.add_edge((3, 2))

graph1_adj_m = graph_to_adj_m(graph1)

pathgraph1 = copy.deepcopy(graph1)
mark_edge(pathgraph1, 3, 2, True, 1)
mark_edge(pathgraph1, 2, 1, False, 1)
mark_edge(pathgraph1, 1, 0, True, 1)
mark_edge(pathgraph1, 0, 3, False, 1)






##
# aumentation cyrcle
graph2 = RstDiGraph(indegree_serie=[0, 1, 0, 1, 0, 1], outdegree_serie=[1, 0, 1, 0, 1, 0], restriction_set_list=[set([0, 1, 2]), set([3, 4, 5])], crossing_np_array = np.array(([1, 1], [0, 1])))
graph2.add_edge((0,1))
graph2.add_edge((2,5))
graph2.add_edge((4,3))

var_dict2 = {
    0: {'gender': 'm', 'age': 50},
    1: {'gender': 'f', 'age': 50},
    2: {'gender': 'z', 'age': 50},
    3: {'gender': 'z', 'age': 51},
    4: {'gender': 'z', 'age': 51},
    5: {'gender': 'z', 'age': 51},
}

graph2_adj_m =graph_to_adj_m(graph2)


# long cycle
indegree_serie =[0,1,0,2,0,2,0,1]
outdegree_serie=[1,0,2,0,2,0,1,0]
restriction_set_list=[set([0,1,2,3]),set([4,5,6,7])]
crossing_np_array = np.array(([0,0],[0,0]))
graph3 = RstDiGraph(indegree_serie, outdegree_serie, restriction_set_list, crossing_np_array)
graph3emty = RstDiGraph(indegree_serie, outdegree_serie, restriction_set_list, crossing_np_array)

graph3.add_edge((0,1))
graph3.add_edge((2,3))
graph3.add_edge((2,5))
graph3.add_edge((4,3))
graph3.add_edge((4,5))
graph3.add_edge((6,7))
graph_correct(graph3)

# impossible long
indegree_serie =[2,0,2,1,1,2,0]
outdegree_serie=[1,1,0,3,2,0,1]
restriction_set_list=[set([0,1,2,3]),set([4,5,6])]
crossing_np_array = np.array(([0,0],[0,0]))
graph4 = RstDiGraph(indegree_serie, outdegree_serie, restriction_set_list, crossing_np_array)
graph4emty = RstDiGraph(indegree_serie, outdegree_serie, restriction_set_list, crossing_np_array)
graph4.add_edge((1,0))
graph4.add_edge((0,2))
graph4.add_edge((3,0))
graph4.add_edge((3,2))
graph4.add_edge((3,4))
graph4.add_edge((4,3))
graph4.add_edge((4,5))
graph4.add_edge((6,5))
graph_correct(graph4)

# possible not equlibrated:
# long cycle
indegree_serie =[0,2,1,1,1,1,1,1]
outdegree_serie=[2,0,0,3,0,1,1,1]
restriction_set_list=[set([0,1,2,3,4]),set([5,6,7])]
crossing_np_array = np.array(([0,0],[0,0]))
graph5 = RstDiGraph(indegree_serie, outdegree_serie, restriction_set_list, crossing_np_array)
graph5emty = RstDiGraph(indegree_serie, outdegree_serie, restriction_set_list, crossing_np_array)
graph5.add_edge((0,1))
graph5.add_edge((0,2))
graph5.add_edge((3,1))
graph5.add_edge((3,4))
graph5.add_edge((3,5))
graph5.add_edge((5,3))
graph5.add_edge((6,7))
graph5.add_edge((7,6))
graph_correct(graph5)



import unittest
import pytest
import numpy as np
from ugd import graph_hyp_test
from test.test_resources.plain_graph_integration import adj_m_1, var_dict_1


def my_stat(adm, var_dict):
    return 5


class UserInterface(unittest.TestCase):

    def test_empty_graph(self):
        adj_m = np.zeros((4, 4))
        with pytest.raises(ValueError) as e_info:
            _ = graph_hyp_test(adj_m=adj_m, var_dict=var_dict_1, test_variable=('gender', 'm', 'f'),
                                                mixing_time=100, anz_sim=100, show_polt=False)


    def test_only_matrix(self):
        _ = graph_hyp_test(adj_m_1, mixing_time=10)
        # should not rise an error
        assert True


    def test_spelling_error(self):
        with pytest.raises(ValueError) as e_info:
            _ = graph_hyp_test(adj_m=adj_m_1, var_dict=var_dict_1, test_variable=('genderr', 'm', 'f'))


    def test_no_node_with_testvalue(self):
        with pytest.raises(ValueError) as e_info:
            _ = graph_hyp_test(adj_m=adj_m_1, var_dict=var_dict_1, test_variable=('gender', 'male', 'f'))


    def test_no_node_with_testvalue2(self):
        with pytest.raises(ValueError) as e_info:
            _ = graph_hyp_test(adj_m=adj_m_1, var_dict=var_dict_1, test_variable=('gender', 'm', 'female'))


    def test__with_only_stat(self):
        _ = graph_hyp_test(adj_m=adj_m_1, stat_f=my_stat, mixing_time=10)
        assert True


    def test__with_only_stat_and_vardict(self):
        _ = graph_hyp_test(adj_m=adj_m_1, var_dict=var_dict_1, stat_f=my_stat, mixing_time=10)
        assert True


    def test_controlls_in_vardict(self):
        _ = graph_hyp_test(adj_m=adj_m_1, var_dict=var_dict_1,
                                            stat_f=my_stat, mixing_time=10)
        assert True


    def test_with_control_misspeled(self):
        with pytest.raises(ValueError) as e_info:
            _ = graph_hyp_test(adj_m=adj_m_1, var_dict=var_dict_1, controls=['genderr'], mixing_time=10)


    def test_matrix_with_vardict(self):
        _ = graph_hyp_test(adj_m_1, var_dict=var_dict_1, mixing_time=10)
        assert True



from scipy.stats import logistic
import numpy as np
from util.graph_adj_m_util import clear_selfloops, get_pot_triad_m


def pos_weight_f(value):
    return logistic.pdf(value) / logistic.cdf(value)


def neg_weight_f(value):
    return logistic.pdf(value) / (1-logistic.cdf(value))


def create_lcl_mst_pwf_statistic_from_thresholds(threholds_m, n, edge_util_function = None):
    """
    :param threholds_m: Thresholds_m are te thresholds per edge denoted in the paper with $t$
    :param n: number of nodes
    :param edge_util_function: function which maps a graph represented in the adjacency matrix to a matrix of the same
                               shape representing the additional utility of the agent form forming the edge
    :return:
    """
    if edge_util_function == None:
        print("Waring: No edge utility function provided, using the utility equal to the number of triads closed")
        edge_util_function = get_pot_triad_m

    # for each link
    pos_weight_m = np.zeros((n, n))
    neg_weight_m = np.zeros((n, n))
    for i in range(n):
        for j in range(n):
            if i == j:  # self-loops not counted
                continue
            pos_weight_m[i, j] += pos_weight_f(threholds_m[i, j])
            neg_weight_m[i, j] += neg_weight_f(threholds_m[i, j])

    def stat(adj_m, vardict = None): # vardict is given as optional in order to be conform with the ugd API
        pot_triad_m = edge_util_function(adj_m=adj_m)  # number of triads for a link
        positive_value = np.sum(np.multiply(np.multiply(pot_triad_m, adj_m), pos_weight_m))
        neg_adj_m = clear_selfloops(1 - adj_m)
        negative_value = np.sum(np.multiply(np.multiply(pot_triad_m, neg_adj_m), neg_weight_m))
        value = positive_value - negative_value
        return value
    return stat


def get_edge_treholds(node_dict):
    n = node_dict.__len__()

    edge_trehold_m = np.zeros((n,n))

    for i in range(n):
        for j in range(n):
            edge_trehold_m[i,j] = node_dict[i][0]+node_dict[j][1] # here in out affection determined (dimentions)
    return edge_trehold_m







'''
This file is a implementation of the algorithm Markov Draw form the paper:

Uniform Sampling of Graphs with Fixed Degree Sequence under Partition Constraints

'''
import numpy as np

'''costume functions'''
from ugd.help_function.util import rand_choise
from ugd.schlaufen_construction.schlaufen import add_random_schlaufe
from ugd.markov_walk.markov_walk_util import switch_cycles, del_marks, update_violation_matrix
from ugd.markov_walk.constraint_violation_check import no_violations
from ugd.help_function.controll_functions_graph import full_graph_correct

def markov_walk(graph, mixing_time):
    '''
    :param graph: RstDiGraph or RstGraph class
    :param mixing_time: integer
    :return: randomly modified  RstDiGraph or RstGraph class, if mixing time is sufficiency big it is a uniform sample
    '''

    # loop over random alteration (step 1-4, in the paper)
    for i in range(mixing_time):
        graph = do_random_step(graph)
    return graph


def do_random_step(graph):
    q = 0.05
    if rand_choise(q):  # self loop (step 2)
        return graph
    else:
        graph, is_feasible, start_nodes, switch_cycle_nodes, active_startnodes = create_schlaufen_sequence(graph)
        # switching
        if is_feasible:  # feasable Schlaufen sequence found
            switch_cycles(graph, switch_cycle_nodes, active_startnodes)
            del_marks(graph, start_nodes)

            return graph
        else:  # self loop
            del_marks(graph, start_nodes)
            return graph


def create_schlaufen_sequence(graph):
    # finding schlaufen sequence (step 3)
    k =graph.restriction_set_list.__len__()
    violation_matrix = np.zeros((k,k))
    ad_schlaufen = True
    start_nodes = []
    cycle_start_nodes = []
    active_cyclenodes = []

    ind = 0
    while ad_schlaufen:
        start_node, cycle_start_node, active_cyclenode = add_random_schlaufe(graph, ind)
        # book keeping on marked Schlaufen
        start_nodes.append(start_node)
        cycle_start_nodes.append(cycle_start_node)
        active_cyclenodes.append(active_cyclenode)
        violation_matrix = update_violation_matrix(graph, cycle_start_node, active_cyclenode, ind, violation_matrix)

        # find a feasibel combination of Schlaufen
        is_feasible = no_violations(violation_matrix=violation_matrix)
        if is_feasible:
            return graph, is_feasible, start_nodes, cycle_start_nodes, active_cyclenodes
        else:
            if rand_choise(0.5):  # note feasible, random interruption
                return graph, is_feasible, start_nodes, cycle_start_nodes, active_cyclenodes
        ind += 1




from test.test_resources.graphs_two_restriction_sets import pathgraph1
from ugd.markov_walk.markov_walk_util import switch_cycles
import unittest


class TestMarkovWalk(unittest.TestCase):

    def test_swich_cycles(self):
        graph = switch_cycles(graph=pathgraph1, cycle_nodes= [None, 3], active_startnodes=[None, True])
        # check if (3,0 ) is in graph
        assert graph.nodes[3].outnodes == set([0])



import numpy as np
import random
#
# creates the original graph for estimaton container
time_layer_list = [2018,2015,2012,2009]
def createGraph(imputPaperList, timeLayerList, firstJornalName):

    # preparation
    var_dict = {}
    has_dict = {}
    i = 0
    for paperNode in imputPaperList:
        if timerestriction_violation(paperNode):
            continue
        has_dict[paperNode['id']] = i
        i += 1
    adj_m = np.zeros((has_dict.__len__(), has_dict.__len__()))

    i = 0
    for paperNode in imputPaperList:
        if timerestriction_violation(paperNode):
            continue
        # Variable filling up
        inerdict = {}
        inerdict['journal'] = paperNode['journal']
        inerdict['timelayer']= isInTimelayer(paperNode['year'],time_layer_list)
        inerdict['rand'] = np.random.binomial(1, 0.5)
        var_dict[i] = inerdict

        for outhash in paperNode['references']:
            if outhash in has_dict:
                j = has_dict[outhash]
            else:
                continue # papers out of the timerange
            if j==i:
                print('selfcitaion')
                continue
            outnode = getNodeFromHash(outhash, imputPaperList)
            if isInTimelayer(paperNode['year'],time_layer_list)>=isInTimelayer(outnode['year'],time_layer_list):
                print('citation in timelayer violation')
                continue
            adj_m[i,j] = 1
        i +=1
    return adj_m, var_dict

def timerestriction_violation(paperNode):
    if int(paperNode['year']) < 2009:
        return True
    if int(paperNode['year']) >= 2018:
        return True
    return False



def isInTimelayer(year,timeLayerList):
    year = int(year)
    layer = None
    for i, iterYear in enumerate(timeLayerList[0:-1]):
        if year < timeLayerList[i] and year >= timeLayerList[i+1]:
            layer = i
            return layer
    return layer

def getNodeFromHash(has,imputPaperList ):
    for node in imputPaperList:
        if node['id']==has:
            return node



import numpy as np
# from ugd.markov_walk.markov_walk_util import is_feasible , set_elm_to_non_exept
from itertools import combinations

''' 
no_violations checks whether all partition constraint are fulfilled.

no_violation is a interface, which can be customized. 

Examples are if you are only interested in the restriction of edges between two specific groups. In this case only the 
specific element in the matrix has to be checked.

Another example is fixing the number of triads in the graph.

'''


def no_violations(violation_matrix, graph = None):
    return np.all(violation_matrix==0)







import json
from test.test_resources.citation_example_graph import adj_m_cit, var_dict_citation
from ugd.high_level_interface.draw_simulations import  digraph_hyp_test

global DEBUG  # Needed to modify global copy of globvar
DEBUG = True
'''actual estimation'''
graphs, stats_list = digraph_hyp_test(adj_m=adj_m_cit, var_dict = var_dict_citation, test_variable= ('paper','B','A'), controlls=['timelayer','random'],  anz_sim=1000, show_polt=True)




from setuptools import setup, find_packages, Extension
import os


modulelist = []
for path, subdirs, files in os.walk('./ugd'):
    for name in files:
          modulelist.append(os.path.join(path, name))

try:
    with open("readme.md", "r") as fh:
        long_description = fh.read()
except:
    long_description = 'Drawing uniformly graphs under partition constraints (Partition Adjacency Matrix). Commonly used for network testing.'


setup(name='ugd',
      version='0.6.1',
      description= 'Drawing uniformly graphs under partition constraints (Partition Adjacency Matrix). Commonly used for network testing.',
      long_description=long_description,
      long_description_content_type='text/markdown',
    classifiers=[
        "Programming Language :: Python :: 3",
    ],
    install_requires=[
          'numpy',
          'matplotlib',
          'statsmodels',
          'pandas'
      ],
      author='Andrin Pelican',
      author_email='pelicanandrin@gmail.com',
      packages=find_packages()
     )



'''

This file implements the Schlaufen Detection algorithm in the paper:

Uniform Sampling of Graphs with Fixed Degree Sequence under Partition Constraints

'''

from ugd.schlaufen_construction.schlaufen_construction_util import mark_edge, mark_node, cycle_found, random_draw, is_schlaufe_type_3
from ugd.help_function.util import rand_element_of_set, del_nodes_mark


def add_plain_random_schlaufe(graph, schleifen_number):
    # random draw of initial node (step 1)
    start_node = rand_element_of_set(range(graph.node_number))

    cycle_node = None
    active_cycle_node = None
    working_node = start_node
    is_active = True
    is_schlaufe = False

    while not (is_schlaufe):


        if is_schlaufe_type_3(graph, working_node, is_active, start_node):
            is_schlaufe = True # schlaufe of type 3:

        else:

            mark_node(graph, working_node, is_active)
            found, out_node = random_draw(graph, working_node, is_active)  # step 2 or 4, depending on is_active

            if not(found):
                is_schlaufe = True  # schlaufe of type 2:
            else:
                # follow edge
                mark_edge(graph, working_node, out_node, is_active, schleifen_number)
                working_node = out_node

                if cycle_found(graph, out_node, is_active):  # step 3 or 5
                    active_cycle_node = not is_active  # out node is one step ahead
                    cycle_node = out_node
                    is_schlaufe = True # schlaufe of type 1:
                else:
                    is_active = not is_active

    del_nodes_mark(graph, start_node)  # but don't delete marked edges
    return start_node, cycle_node, active_cycle_node



'''
mcd:

- digraph_hyp_test

'''

from .high_level_interface.user_interface import digraph_hyp_test,graph_hyp_test
from .locally_most_powerful_stat.lcl_mst_pwf_stat_statsmodels import make_lcl_mst_pwf_stat





import numpy as np

from ugd.help_function.util import graph_to_adj_m
from ugd.model.digraph import RstDiGraph

# long cycle           [0,1,2,3,4,5,6,7,8,9,0,1,2,3,4,5,6,7,8]
indegree_serie_cit   = [0,0,0,0,0,0,0,0,0,2,0,0,1,1,2,1,1,2,0]
outdegree_serie_cit  = [0,2,0,1,0,1,3,0,0,1,2,0,0,0,0,0,0,0,0]
restriction_set_list_cit = [set([0,1,2,3,4,5,6]),set([7,8,9,10,11,12]),set([13,14,15,16,17,18])]
crossing_np_array_mr_cit = np.array(([0,3,4],[0,0,3],[0,0,0]))
graph_cit = RstDiGraph(indegree_serie_cit, outdegree_serie_cit, restriction_set_list_cit, crossing_np_array_mr_cit)

var_dict_citation = \
{
    0 : {'timelayer': 0, 'paper': 'A'},
    1 : {'timelayer': 0, 'paper': 'A'},
    2 : {'timelayer': 0, 'paper': 'A'},
    3 : {'timelayer': 0, 'paper': 'A'},
    4 : {'timelayer': 0, 'paper': 'A'},
    5 : {'timelayer': 0, 'paper': 'B'},
    6 : {'timelayer': 0, 'paper': 'B'},
    7 : {'timelayer': 1, 'paper': 'A'},
    8 : {'timelayer': 1, 'paper': 'A'},
    9 : {'timelayer': 1, 'paper': 'A'},
    10: {'timelayer': 1, 'paper': 'B'},
    11: {'timelayer': 1, 'paper': 'B'},
    12: {'timelayer': 1, 'paper': 'B'},
    13: {'timelayer': 2, 'paper': 'A'},
    14: {'timelayer': 2, 'paper': 'A'},
    15: {'timelayer': 2, 'paper': 'A'},
    16: {'timelayer': 2, 'paper': 'B'},
    17: {'timelayer': 2, 'paper': 'B'},
    18: {'timelayer': 2, 'paper': 'B'},
}

graph_cit.add_edge((1, 13))
graph_cit.add_edge((1, 16))
graph_cit.add_edge((3, 15))
graph_cit.add_edge((5, 9))
graph_cit.add_edge((6, 9))
graph_cit.add_edge((6, 17))
graph_cit.add_edge((6, 12))
graph_cit.add_edge((9, 14))
graph_cit.add_edge((10, 14))
graph_cit.add_edge((10, 17))


adj_m_cit = graph_to_adj_m(graph_cit)





from copy import deepcopy

import numpy as np

from ugd.model import RstGraph

# long cycle
indegree_serie_mr_0 = [1,1,1]
outdegree_serie_mr_0 =[1,1,1]
restriction_set_list_mr_0 = [set([0]),set([1]),set([2])]
crossing_np_array_mr_0 = np.array(([0,1,0],[0,0,1],[1,0,0]))
graph_mr_0_emty = RstGraph(indegree_serie_mr_0, outdegree_serie_mr_0,restriction_set_list_mr_0,crossing_np_array_mr_0 )



# long cycle
indegree_serie_mr_1 = [1,1,1,1,1,1]
outdegree_serie_mr_1 =[1,1,1,1,1,1]
restriction_set_list_mr_1 = [set([0,1]),set([2,3]),set([4,5])]
crossing_np_array_mr_1 = np.array(([0,1,0],[0,0,1],[1,0,0]))

graph_mr_1_emty = RstGraph(indegree_serie_mr_1, outdegree_serie_mr_1,restriction_set_list_mr_1,crossing_np_array_mr_1 )
graph_mr_1 = RstGraph(indegree_serie_mr_1, outdegree_serie_mr_1,restriction_set_list_mr_1,crossing_np_array_mr_1 )
graph_mr_1.add_edge((5,0))
graph_mr_1.add_edge((0,2))
graph_mr_1.add_edge((2,3))
graph_mr_1.add_edge((3,1))
graph_mr_1.add_edge((1,4))
graph_mr_1.add_edge((4,5))
graph_mr_1.update_crossing_surplus_array()

graph_mr_2 = deepcopy(graph_mr_1_emty)
graph_mr_2.add_edge((5,0))
graph_mr_2.add_edge((0,2))
graph_mr_2.add_edge((2,3))
graph_mr_2.add_edge((3,5))
graph_mr_2.add_edge((1,4))
graph_mr_2.add_edge((4,1))
graph_mr_2.working_node = 5
graph_mr_2.fixed_to_nodes = set([0])
graph_mr_2.update_crossing_surplus_array()


'''
Validates the user input, to check whether the API requirements are fulfilled and whether the input graph is correct.
It there is a violation it throws an error with a hint what is violated.
Depending on the input, the default parameters are substituted.
'''

import numpy as np
from ugd.high_level_interface.construct_node_partition import constr_partition
from ugd.help_function.util_statistic import get_default_stat, crossarrow_count
import warnings

''' Validation '''


def validate_input(adj_m, anz_sim, mixing_time, var_dict, stat_f, controlls, test_variable, is_directed):
    # throws error if violated
    adj_m = validate_adj_matrix(adj_m, is_directed)
    anz_sim = validate_pos_int(anz_sim)
    if not (mixing_time == None):
        mixing_time = validate_pos_int(mixing_time)
    var_dict = validate_var_dict(var_dict, adj_m.shape[0])
    stat_f = validate_stat_f(stat_f, adj_m=adj_m, var_dict=var_dict)
    controlls = validate_controls(controlls, var_dict)
    test_variable = validate_test_variable(test_variable, var_dict)


def validate_adj_matrix(adj_m, is_directed):
    if not (isinstance(adj_m, np.ndarray)):
        raise ValueError("Adjacency matrix must be a numpy array.")

    n = adj_m.shape[0]
    if not (adj_m.shape[0] == adj_m.shape[1]):
        ValueError('Adjacency matrix must be quadratic (n x n).')

    for i in range(n):
        for j in range(n):
            if not (adj_m[i, j] == 0) and not (adj_m[i, j] == 1):
                raise ValueError('Matrix is only allowed to have 0, 1 entries (no double arrow).')

            if not (is_directed):  # the plain graph case
                if not (adj_m[j, i] == adj_m[i, j]):
                    raise ValueError('For plain graph adjacency matrix must be symmetric.')
    if adj_m.sum() == 0:
        raise ValueError('Adjacency matrix must have at least one nonzero entry, empty graph not allowed.')

    for i in range(n):
        if adj_m[i, i] == 1:
            raise ValueError('Diagonal entries of matrix must be 0, (no self loops).')
    return adj_m


def validate_pos_int(anz_sim):
    try:
        anz_sim = int(anz_sim)
    except ValueError:
        print("anz_sim must be an integer.")
    if anz_sim <= 0:
        raise ValueError("anz_sim must be positive.")
    return anz_sim


def validate_var_dict(var_dict, n):
    if var_dict == None:
        return True
    if not (isinstance(var_dict, dict)):
        raise ValueError("var_dict must be a dictionary.")

    if not (n == var_dict.__len__()):
        raise ValueError(
            'var_dict must be a dictionary with primary key the integer 1..n, where n is the number of nodes.')

    for i in range(n):
        if not (i in var_dict):
            raise ValueError(
                'var_dict must be a dictionary with primary key the integer 1..n, where n is the the number of nodes.')
        if not (isinstance(var_dict[i], dict)):
            raise ValueError(
                "Values of var_dict must be a dictionary, with the variable name as key and the variable value as value")
    return var_dict


def validate_stat_f(stat_f, adj_m, var_dict):
    if not (stat_f == None):
        try:
            a = stat_f(adj_m, var_dict)
            try:
                b = int(a)
            except:
                raise ValueError('stat_f must be a function of form stat_f(adj_m, var_dict), and return a number.')
        except:
            raise ValueError('stat_f must be a function of form stat_f(adj_m, var_dict), and return a number.')
    return stat_f


def validate_controls(controls, var_dict):
    if controls == None:
        return controls
    if not (isinstance(controls, list)):
        raise ValueError('controls must be a list of variable names.')
    for i in range(var_dict.__len__()):  # see whether the controls are present in all the values of the var_dict
        for control in controls:
            if not (control in var_dict[i]):
                raise ValueError(
                    'controls must be a list of variable names, the variable names key in the all the values of var_dict \n'
                    'controls[i] in var_dict[i], for i in {0,..,n-1}')
    return controls


def validate_test_variable(test_variable, var_dict):
    if test_variable == None:
        return test_variable

    if not (isinstance(test_variable, tuple)):
        raise ValueError(' test_variable must be a tuple with 3 entries, \n'
                         ' first variable name of interest \n'
                         ' second: value of the variable form which the arrow depart \n'
                         ' third: value of the variable go which the arrow go \n'
                         )
    first_entry = False
    second_entry = False
    for i in range(var_dict.__len__()):
        if not (test_variable[0] in var_dict[i]):
            raise ValueError('test_variable must be a tuple with 3 entries, \n' +
                             'first:   variable name of interest \n' +
                             'second:  value of the variable from which the link depart \n' +
                             'third:   value of the variable go which the link go \n' +
                             'The variable name was not in the variable dict of node ', i)

        # Verify whether there are nodes with the values to test for
        if (test_variable[1] == var_dict[i][test_variable[0]]):
            first_entry = True
        if (test_variable[2] == var_dict[i][test_variable[0]]):
            second_entry = True
    if not (first_entry):
        raise ValueError('There are no nodes with the value: ', test_variable[1],
                         ' which is on the first place in the test tuple')
    if not (second_entry):
        raise ValueError('There are no nodes with the value: ', test_variable[2],
                         ' which is on the second place in the test tuple')
    return test_variable


def validate_nodesetpartition(nodesetpartition, n):
    if not (isinstance(nodesetpartition, list)):
        raise ValueError(" node partiton must be a list of sets")
    for nodest in nodesetpartition:
        if not (isinstance(nodest, set)):
            raise ValueError(" node partiton must be a list of sets")
    set_list = nodesetpartition
    total_Set = set()
    total = 0
    for sub_set in set_list:
        total_Set = total_Set.union(sub_set)
        total += sub_set.__len__()
    if not (total == total_Set.__len__()):
        raise ValueError('restriction sets are not disjoint')
    if not (total_Set == set(range(n))):
        raise ValueError('nodes in set are not the integers {0,..n-1}')
    return nodesetpartition


''' Substitution '''


def parse_input(stat_f, test_variable, controlls, var_dict, adj_m):
    # substitution input (which can be of different shape into standard shape
    if not (stat_f == None) and not (test_variable == None):
        warnings.warn('test_variable and stat_f where given, only stat_f is evaluated')
    # input substitution
    if stat_f == None:
        stat_f = get_default_stat(adj_m=adj_m)
        if not (test_variable == None):
            def crossarrow_stat(adj_m, v_dict):
                return crossarrow_count(adj_m, v_dict, test_variable)

            stat_f = crossarrow_stat

    # creation of set partition
    nodesetpartition = constr_partition(controls=controlls, var_dict=var_dict, adj_m=adj_m)
    validate_nodesetpartition(nodesetpartition, adj_m.shape[0])
    return stat_f, nodesetpartition
